""" base classes for creating SVG files """
from html import escape
from pathlib import Path
from typing import Optional


class SvgElement:
    """ general class for describing an element in an SVG and transforming it into
    its XML representation for saving it """
    def __init__(self, tag: str,
                 attributes: Optional[dict[str, str]] = None,
                 content: Optional[str] = None,
                 classes: Optional[list[str]] = None,
                 ):
        self._tag = tag
        self._attributes = attributes or {}
        self._content = content
        self._add_classes(classes)

    @property
    def tag(self) -> str:
        """ the element's tag as a string """
        return self._tag

    def _add_classes(self, classes: Optional[list[str]] = None) -> None:
        """ add new classes to the element's attributes """
        if classes is None or len(classes) == 0:
            return
        if 'class' not in self._attributes:
            self._attributes['class'] = ' '.join(classes)
            return
        for class_name in classes:
            if class_name in self._attributes['class']:
                continue
            self._attributes['class'] += ' ' + class_name

    @property
    def attributes(self) -> dict[str, str]:
        """ dictionary of the element's attributes """
        return self._attributes

    @property
    def classes(self) -> list[str]:
        """ list of the element's classes """
        if 'class' in self._attributes:
            return self._attributes['class'].split(' ')
        return []

    @property
    def content(self) -> Optional[str]:
        """ the element's content as a string """
        return self._content

    def __str__(self) -> str:
        svg_element = f'<{self.tag}'
        for key, value in self.attributes.items():
            svg_element += f' {key}="{escape(value)}"'
        if self.content is not None:
            svg_element += f'>{self.content}</{self.tag}>'
        else:
            svg_element += ' />'
        return svg_element


class SVG:
    """ representation of an SVG file used to collect the contained elements
    and save them into a .svg along with the necessary meta-data
    """
    def __init__(self, width: int, height: int,
                 style: Optional[dict[str, dict[str, str]]] = None,
                 elements: Optional[list[SvgElement]] = None,
                 definitions: Optional[list[SvgElement]] = None):
        self.width = width
        self.height = height
        self.style = style or {}
        self.elements = elements or []
        self.defs = definitions or []

    @property
    def header(self) -> str:
        """ first lines of the .svg file """
        width, height = int(self.width), int(self.height)
        view_x, view_y = 0, 0
        lines = [
            '<?xml version="1.0" encoding="UTF-8"?>',
            '<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"',
            f'     width="{width}" height="{height}" viewBox="{view_x} {view_y} {width} {height}">',
        ]
        return '\n'.join(lines) + '\n'

    @property
    def style_section(self) -> str:
        """ style section lines of the .svg file """
        style_section = '<style>\n'
        indent = ' ' * 4
        for selector, props in self.style.items():
            style_section += f'{selector} {{\n'
            for name, value in props.items():
                style_section += f'{indent}{name}: {value};\n'
            style_section += '}\n'
        style_section += '</style>\n'
        return style_section

    @property
    def defs_section(self) -> str:
        """ definition section lines of the .svg file """
        defs_section = ''
        indent = ' ' * 4
        if len(self.defs) > 0:
            defs_section += '<defs>\n'
            defs_section += ''.join(indent + str(element) + '\n'
                                    for element in self.defs)
            defs_section += '</defs>\n'
        return defs_section

    @property
    def element_section(self) -> str:
        """ main elements section lines of the .svg file """
        element_section = ''.join(str(element) + '\n'
                                  for element in self.elements)
        return element_section

    @property
    def footer(self) -> str:
        """ last lines of the .svg file """
        return '</svg>\n'

    @property
    def full(self) -> str:
        """ the full raw .svg file """
        full = self.header
        full += self.style_section
        full += self.defs_section
        full += self.element_section
        full += self.footer
        return full

    def save_as(self, file_path: Path) -> None:
        """ save the SVG under given file path """
        with open(file_path, 'w', encoding='utf-8') as out_file:
            out_file.write(self.full)



""" classes and functions for easier calculations on datetimes and coordinates """
import calendar
from collections.abc import Iterable
from datetime import datetime

from svg_timeline.geometry import Vector


class TimeGradient:
    """ class for the transfer of dates to canvas coordinates and back """
    def __init__(self, source: Vector, target: Vector, start_date: datetime, end_date: datetime):
        """
        :param source: the point on the canvas that correspond to the start of the gradient
        :param target: the point on the canvas that correspond to the end of the gradient
        :param start_date: the datetime that corresponds to the start of the canvas_vector
        :param end_date: the datetime that corresponds to the end of the canvas_vector
        """
        self._source = source
        self._target = target
        self._start_date = start_date
        self._end_date = end_date

    @property
    def source(self) -> Vector:
        """ the point on the canvas that correspond to the start of the gradient """
        return self._source

    @property
    def target(self) -> Vector:
        """ the point on the canvas that correspond to the end of the gradient """
        return self._target

    @property
    def start_date(self) -> datetime:
        """ the datetime that corresponds to the start of the canvas_vector """
        return self._start_date

    @property
    def end_date(self) -> datetime:
        """ the datetime that corresponds to the end of the canvas_vector """
        return self._end_date

    def coord_to_date(self, coord: Vector) -> datetime:
        """ transform an absolute position on the canvas into a date """
        return self.relative_to_date(self.coord_to_relative(coord=coord))

    def coord_to_relative(self, coord: Vector) -> float:
        """ transform an absolute position on the canvas
        into a relative position on the timeline
        """
        # Transform coordinates so that the timeline start is at (0, 0).
        # (simplifies the following calculations)
        coord_x = coord.x - self._source.x
        coord_y = coord.y - self._source.y
        end_x = self._target.x - self._source.x
        end_y = self._target.y - self._source.y
        # Given a scalar factor 'a', minimize the length of vector 'coord - a * end'.
        # 'a' then describes the relative position on this timeline with the
        # shortest distance to the given coordinates.
        # Solved analytically, this gives:
        numerator = coord_x * end_x + coord_y * end_y
        denominator = end_x**2 + end_y**2
        a = numerator / denominator
        return a

    def date_to_coord(self, date: datetime) -> Vector:
        """ transform a date into a position on the canvas """
        return self.relative_to_coord(self.date_to_relative(date=date))

    def date_to_relative(self, date: datetime) -> float:
        """ transform a date into a relative position on the timeline """
        self_delta = self._end_date - self._start_date
        date_delta = date - self.start_date
        return date_delta / self_delta

    def relative_to_coord(self, relative_position: float) -> Vector:
        """ transform a relative position on the timeline
        into an absolute position on the canvas
        """
        delta = self._target - self._source
        scaled_vector = self._source + (relative_position * delta)
        return scaled_vector

    def relative_to_date(self, relative_position: float) -> datetime:
        """ transform a relative position on the timeline into a date """
        delta = self._end_date - self._start_date
        return self._start_date + relative_position * delta


class TimeSpacing:
    """ base class for semantic datetime spacing within a given range """
    def __init__(self, start_date: datetime, end_date: datetime):
        if not start_date < end_date:
            raise ValueError("start date needs to be smaller than end date")
        self._start_date = start_date
        self._end_date = end_date

    @property
    def start_date(self) -> datetime:
        """ the datetime that corresponds to the start of the time range """
        return self._start_date

    @property
    def end_date(self) -> datetime:
        """ the datetime that corresponds to the end of the time range """
        return self._end_date

    @property
    def labels(self) -> list[str]:
        """ Tic labels
        :return list of tic labels as strings
        """
        raise NotImplementedError

    @property
    def dates(self) -> list[datetime]:
        """ Positions of the tics
        :return list of tic positions as datetime objects
        """
        raise NotImplementedError


class YearBasedTimeSpacing(TimeSpacing):
    """ base class to return one entry per X years """
    _base = 1

    @property
    def _range(self) -> Iterable:
        first = (self._start_date.year // self._base + 1) * self._base
        last = (self._end_date.year // self._base) * self._base
        return range(first, last + self._base, self._base)

    @property
    def dates(self) -> list[datetime]:
        dates = [datetime(year=year, month=1, day=1)
                 for year in self._range]
        return dates

    @property
    def labels(self) -> list[str]:
        labels = [str(value) for value in self._range]
        return labels


class TimeSpacingPerMillennia(YearBasedTimeSpacing):
    """ return one entry per 1000 years
    Note: ISO 8601 only allows years between 0000 and 9999
    """
    _base = 1000


class TimeSpacingPerCentury(YearBasedTimeSpacing):
    """ return one entry per 100 years """
    _base = 100


class TimeSpacingPerDecade(YearBasedTimeSpacing):
    """ return one entry per 10 years """
    _base = 10


class TimeSpacingPerYear(YearBasedTimeSpacing):
    """ return one entry per year """
    _base = 1


class TimeSpacingPerMonth(TimeSpacing):
    """ return one entry per month """
    @property
    def dates(self) -> list[datetime]:
        year = self.start_date.year
        month = self.start_date.month + 1
        dates = []
        while True:
            if month > 12:
                year += 1
                month = 1
            date = datetime(year=year, month=month, day=1)
            if date > self.end_date:
                break
            dates.append(date)
            month += 1
        return dates

    @property
    def labels(self) -> list[str]:
        labels = [calendar.month_abbr[date.month] for date in self.dates]
        return labels


class TimeSpacingPerDay(TimeSpacing):
    """ return one entry per day """
    @property
    def dates(self) -> list[datetime]:
        year = self.start_date.year
        month = self.start_date.month
        _, n_days = calendar.monthrange(year, month)
        day = self.start_date.day + 1
        dates = []
        while True:
            if day > n_days:
                month += 1
                day = 1
                if month > 12:
                    year += 1
                    month = 1
                    _, n_days = calendar.monthrange(year, month)
            date = datetime(year=year, month=month, day=day)
            if date > self.end_date:
                break
            dates.append(date)
            day += 1
        return dates

    @property
    def labels(self) -> list[str]:
        labels = [str(date.day) for date in self.dates]
        return labels



""" high level timeline API classes """
from datetime import datetime
from pathlib import Path
from typing import Optional

from svg_timeline.geometry import Vector
from svg_timeline.style import Defaults, DEFAULT_CSS, ClassNames
from svg_timeline.svg import SVG
from svg_timeline.svg_primitives import Rectangle, Line, Text, Circle, Image
from svg_timeline.time_calculations import TimeGradient, TimeSpacing


class TimelinePlot:
    """ representation of a timeline plot
    dates, timespans etc. can be added to this timeline via method calls
    """
    def __init__(self, start_date: datetime, end_date: datetime,
                 time_spacing: TimeSpacing, minor_tics: Optional[TimeSpacing] = None,
                 size: tuple[int, int] = (800, 600)):
        self._width, self._height = size
        self._svg = SVG(self._width, self._height, style=DEFAULT_CSS)

        y = Defaults.arrow_y_position * self._height
        x1 = Defaults.arrow_x_padding * self._width
        x2 = (1 - Defaults.arrow_x_padding) * self._width
        self._time = TimeGradient(source=Vector(x1, y), target=Vector(x2, y),
                                  start_date=start_date, end_date=end_date)
        self._tics = time_spacing
        self._tics_minor = minor_tics
        self._add_timeline()

    def _add_timeline(self):
        line = Line(self._time.source, self._time.target, classes=[ClassNames.TIMEAXIS])
        self._svg.elements.append(line)
        timeline_delta = self._time.target - self._time.source
        tic_delta = 10 * timeline_delta.orthogonal()
        for date, label in zip(self._tics.dates, self._tics.labels):
            tic_base = self._time.date_to_coord(date)
            tic_end = tic_base + tic_delta
            text_start = tic_base + 1.5 * tic_delta
            self._svg.elements.append(Line(source=tic_base, target=tic_end, classes=[ClassNames.MAJOR_TICK]))
            self._svg.elements.append(Text(text_start, label, classes=[ClassNames.MAJOR_TICK]))
        if self._tics_minor is None:
            return
        for date in self._tics_minor.dates:
            tic_base = self._time.date_to_coord(date)
            tic_end = tic_base + 0.5 * tic_delta
            self._svg.elements.append(Line(source=tic_base, target=tic_end, classes=[ClassNames.MINOR_TICK]))

    def add_event(self, date: datetime, text: str,
                  lane: int = 1, classes: Optional[list[str]] = None):
        """ Add an event to the timeline that happened at a single point in time """
        classes = classes or []
        classes += [ClassNames.EVENT]
        event_base = self._time.date_to_coord(date)
        event_end = self.__to_lane_point(date, lane=lane)
        text_coord = self.__to_lane_point(date, lane=(lane+0.5 if lane >= 0 else lane-0.5))
        self._svg.elements += [
            Line(source=event_base, target=event_end, classes=classes),
            Circle(center=event_end, radius=Defaults.event_dot_radius, classes=classes),
            Text(text_coord, text, classes=classes),
        ]

    def add_connected_events(self, dates: list[datetime], labels: list[str],
                             classes: Optional[list[Optional[list[str]]]] = None,
                             lane: int = 1,
                             ) -> None:
        """ Add a series of events connected via lines """
        if classes is None:
            classes = [[] for _ in range(len(dates))]
        if not len(dates) == len(labels) == len(classes):
            raise RuntimeError("dates, labels and classes need to be of the same length")
        self._svg.elements += [Line(
            source=self.__to_lane_point(dates[i], lane=lane),
            target=self.__to_lane_point(dates[i+1], lane=lane),
            classes=classes[i],
        ) for i in range(len(dates)-1)]
        self._svg.elements += [Circle(
            center=self.__to_lane_point(dates[i], lane=lane),
            radius=Defaults.event_dot_radius,
            classes=classes[i],
        ) for i, label in enumerate(labels) if label is not None]
        self._svg.elements += [Text(
            coord=self.__to_lane_point(dates[i], lane=(lane+0.5 if lane >= 0 else lane-0.5)),
            text=label,
            classes=classes[i],
        ) for i, label in enumerate(labels) if label is not None]

    def add_image(self, date: datetime, image_path: Path, height: float, width: float,
                  lane: int = 1, classes: Optional[list[str]] = None):
        """ Add an image to the timeline that is associated with a single point in time """
        classes = classes or []
        classes += [ClassNames.IMAGE]
        event_base = self._time.date_to_coord(date)
        event_end = self.__to_lane_point(date, lane=lane)
        image_center_left = event_end + height * self.lane_normal
        image_top_left = image_center_left + width/2 * self.lane_normal.orthogonal(ccw=True)
        self._svg.elements += [
            Line(source=event_base, target=event_end, classes=classes),
            Image(top_left=image_top_left, file=image_path, height=height, width=width, classes=classes),
        ]

    def add_timespan(self, start_date: datetime, end_date: datetime, text: str,
                     lane: int = 1, width: Optional[int] = None, classes: Optional[list[str]] = None):
        """ Add an entry to the timeline that is associated with a certain time span """
        classes = classes or []
        classes += [ClassNames.TIMESPAN]
        width = width or Defaults.timespan_width
        half_width_vector = width/2 * self.lane_normal
        if Defaults.timespan_use_start_stilt:
            on_timeline = self.__to_lane_point(start_date, lane=0)
            bottom_timespan = self.__to_lane_point(start_date, lane=lane) - half_width_vector
            self._svg.elements.append(Line(source=on_timeline, target=bottom_timespan, classes=classes))
        if Defaults.timespan_use_end_stilt:
            on_timeline = self.__to_lane_point(end_date, lane=0)
            bottom_timespan = self.__to_lane_point(end_date, lane=lane) - half_width_vector
            self._svg.elements.append(Line(source=on_timeline, target=bottom_timespan, classes=classes))
        start_corner = self.__to_lane_point(start_date, lane=lane) + half_width_vector
        end_corner = self.__to_lane_point(end_date, lane=lane) - half_width_vector
        middle_date = start_date + (end_date - start_date) / 2
        text_coord = self.__to_lane_point(middle_date, lane=lane)
        self._svg.elements += [
            Rectangle(start_corner, end_corner, classes=classes),
            Text(text_coord, text, classes=classes),
        ]

    def add_title(self, title: str, classes: Optional[list[str]] = None):
        """ Add a title that should be printed above the timeline """
        classes = classes or []
        classes += [ClassNames.TITLE]
        text_coord = Vector(x=int(self._width * Defaults.title_x_position),
                            y=int(self._height * Defaults.title_y_position))
        self._svg.elements += [Text(text_coord, title, classes=classes)]

    @property
    def lane_normal(self) -> Vector:
        """ Normal vector orthogonal to the timeline direction
        This vector is used to calculate the positions of the different lanes.
        """
        return (self._time.target - self._time.source).orthogonal(ccw=True)

    def __to_lane_point(self, date: datetime, lane: float = 1) -> Vector:
        date_coord = self._time.date_to_coord(date)
        lane_point = date_coord + lane * Defaults.lane_width * self.lane_normal
        return lane_point

    def save(self, file_path: Path):
        """ Save an SVG of the timeline under the given file path """
        self._svg.save_as(file_path=file_path)



""" classes for defining different kinds of elements that can be drawn in an SVG """
import math
from mimetypes import guess_type
from base64 import b64encode
from html import escape
from pathlib import Path
from typing import Optional

from svg_timeline.geometry import Vector
from svg_timeline.svg import SvgElement


class Line(SvgElement):
    """ straight line from one point to another """
    def __init__(self, source: Vector, target: Vector, classes: Optional[list[str]] = None):
        super().__init__(tag='path', classes=classes)
        self.source = source
        self.target = target
        self._update_attributes()

    def _update_attributes(self) -> None:
        self._attributes.update({
            'd': f'M{self.source.x},{self.source.y} L{self.target.x},{self.target.y}'
        })


class Text(SvgElement):
    """ text at a fixed position on the canvas """
    def __init__(self, coord: Vector, text: str, classes: Optional[list[str]] = None):
        super().__init__(tag='text', content=escape(text), classes=classes)
        self.coord = coord
        self._update_attributes()

    def _update_attributes(self) -> None:
        self._attributes.update({
            'x': str(self.coord.x),
            'y': str(self.coord.y),
        })


class Rectangle(SvgElement):
    """ rectangle filled with the given color """
    def __init__(self, corner1: Vector, corner2: Vector, classes: Optional[list[str]] = None):
        super().__init__(tag='rect', classes=classes)
        self.corner1 = corner1
        self.corner2 = corner2
        self._update_attributes()

    def _update_attributes(self) -> None:
        self._attributes.update({
            'x': str(min(self.corner1.x, self.corner2.x)),
            'y': str(min(self.corner1.y, self.corner2.y)),
            'width': str(math.fabs(self.corner1.x - self.corner2.x)),
            'height': str(math.fabs(self.corner1.y - self.corner2.y)),
        })


class Circle(SvgElement):
    """ circle filled with the given color """
    def __init__(self, center: Vector, radius: float, classes: Optional[list[str]] = None):
        super().__init__(tag='circle', classes=classes)
        self.center = center
        self.radius = radius
        self._update_attributes()

    def _update_attributes(self) -> None:
        self._attributes.update({
            'cx': str(self.center.x),
            'cy': str(self.center.y),
            'r': str(self.radius),
        })


class Image(SvgElement):
    """ SVG embedding of the image found at the given file path """
    def __init__(self, top_left: Vector, width: float, height: float,
                 file: Path, classes: Optional[list[str]] = None):
        super().__init__(tag='image', classes=classes)
        self.top_left = top_left
        self.width = width
        self.height = height
        self.file = file
        self._update_attributes()

    def _update_attributes(self) -> None:
        mimetype, encoding = guess_type(self.file)
        with open(self.file, 'rb', encoding=encoding) as image_file:
            image_data = b64encode(image_file.read())
        self._attributes.update({
            'x': str(self.top_left.x),
            'y': str(self.top_left.y),
            'width': str(self.width),
            'height': str(self.height),
            'xlink:href': f'data:{mimetype};base64,{image_data.decode()}',
        })



""" Singleton to store styling defaults """
from dataclasses import dataclass
from enum import StrEnum


@dataclass
class __TimelineStyle:
    title_x_position: float = 1/2
    title_y_position: float = 1/17
    title_size_factor: float = 1/15

    lane_width: float = 30

    arrow_y_position: float = 0.9
    arrow_x_padding: float = 0.03

    event_dot_radius: float = 3

    timespan_width: float = 18
    timespan_use_start_stilt: bool = False
    timespan_use_end_stilt: bool = False


Defaults = __TimelineStyle()


class ClassNames(StrEnum):
    """ string constants for all the class names that are commonly used for styling via CSS """
    TITLE = 'title'
    TIMEAXIS = 'time_axis'
    MINOR_TICK = 'minor_tick'
    MAJOR_TICK = 'major_tick'
    EVENT = 'event'
    TIMESPAN = 'timespan'
    IMAGE = 'image'
    WHITE_TEXT = 'white_text'
    COLOR_A = 'color_a'
    COLOR_B = 'color_b'
    COLOR_C = 'color_c'
    COLOR_D = 'color_d'
    COLOR_E = 'color_e'


DEFAULT_CSS = {
    ':root': {
        '--color_a': '#003f5c',
        '--color_b': '#58508d',
        '--color_c': '#bc5090',
        '--color_d': '#ff6361',
        '--color_e': '#ffa600',
    },
    f'path.{ClassNames.COLOR_A}': {'stroke': 'var(--color_a)'},
    f'path.{ClassNames.COLOR_B}': {'stroke': 'var(--color_b)'},
    f'path.{ClassNames.COLOR_C}': {'stroke': 'var(--color_c)'},
    f'path.{ClassNames.COLOR_D}': {'stroke': 'var(--color_d)'},
    f'path.{ClassNames.COLOR_E}': {'stroke': 'var(--color_e)'},
    f'rect.{ClassNames.COLOR_A}, circle.{ClassNames.COLOR_A}': {'fill': 'var(--color_a)'},
    f'rect.{ClassNames.COLOR_B}, circle.{ClassNames.COLOR_B}': {'fill': 'var(--color_b)'},
    f'rect.{ClassNames.COLOR_C}, circle.{ClassNames.COLOR_C}': {'fill': 'var(--color_c)'},
    f'rect.{ClassNames.COLOR_D}, circle.{ClassNames.COLOR_D}': {'fill': 'var(--color_d)'},
    f'rect.{ClassNames.COLOR_E}, circle.{ClassNames.COLOR_E}': {'fill': 'var(--color_e)'},
    'svg': {
        'background': 'white',
    },
    'path': {
        'stroke': 'black',
        'stroke-width': '2pt',
        'fill': 'none',
    },
    'text': {
        'font-size': '10pt',
        'font-family': 'Liberation Sans',
        'fill': 'black',
        'text-anchor': 'middle',
        'dominant-baseline': 'central',
    },
    'circle, rect': {
        'fill': 'black',
    },
    f'text.{ClassNames.TITLE}': {
        'font-size': '20pt',
    },
    f'path.{ClassNames.TIMEAXIS}': {
        'stroke-width': '3pt',
    },
    f'path.{ClassNames.MAJOR_TICK}': {
        'stroke-width': '2pt',
    },
    f'path.{ClassNames.MINOR_TICK}': {
        'stroke-width': '1pt',
    },
    f'path.{ClassNames.EVENT}': {
        'stroke-width': '2pt',
    },
    f'circle.{ClassNames.EVENT}': {
        'radius': '3pt',
    },
    f'path.{ClassNames.TIMESPAN}': {
        'stroke-width': '1pt',
    },
    f'text.{ClassNames.TIMESPAN}': {
        'font-size': '9pt',
    },
    f'path.{ClassNames.IMAGE}': {
        'stroke-width': '2pt',
    },
    f'text.{ClassNames.WHITE_TEXT}': {
        'fill': 'white',
    },
}



""" basic geometry classes to describe canvas points """
import math
from typing import Self

# tolerance on coordinates within which two points are considered equal
COORD_TOLERANCE = 0.000_001


class Canvas:
    """ representation of a rectangular drawing area
    with the origin (0,0) in the top left corner """
    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

    def __repr__(self) -> str:
        return f"Canvas(width={self.width}, height={self.height})"

    def __contains__(self, item) -> bool:
        """ check whether the item is contained within the canvas"""
        if isinstance(item, Vector):
            if (item.x < 0 or item.x > self.width or
                item.y < 0 or item.y > self.height):
                return False
            return True
        raise TypeError(f"__contains__ not defined for type '{type(item)}'")


class Vector:
    """ a vector (or point) within a canvas """
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f"Vector({self.x}, {self.y})"

    def __eq__(self, other) -> bool:
        """ two points are equal, if their coordinates are equal within COORD_TOLERANCE """
        if not isinstance(other, Vector):
            raise TypeError("Can only compare with another CanvasPoint instance")
        return (math.fabs(self.x - other.x) < COORD_TOLERANCE and
                math.fabs(self.y - other.y) < COORD_TOLERANCE)

    def __add__(self, other) -> Self:
        """ component-wise addition with another vector """
        if not isinstance(other, Vector):
            return NotImplemented
        return Vector(self.x + other.x, self.y + other.y)

    def __sub__(self, other) -> Self:
        """ component-wise subtraction with another vector """
        if not isinstance(other, Vector):
            return NotImplemented
        return Vector(self.x - other.x, self.y - other.y)

    def __mul__(self, other) -> Self:
        """ scalar multiplication with an integer or float value """
        if not isinstance(other, (int, float)):
            return NotImplemented
        return Vector(other * self.x, other * self.y)

    def __rmul__(self, other) -> Self:
        """ (see __mul__)"""
        return self.__mul__(other=other)

    def __truediv__(self, other) -> Self:
        """ (see __mul__)"""
        factor = 1/other
        return self.__mul__(other=factor)

    def __rtruediv__(self, other) -> Self:
        """ dividing a value by a vector is not possible """
        return NotImplemented

    @property
    def mag(self) -> float:
        """ the vector magnitude (length) according to the euclidian norm """
        norm = math.sqrt(self.x**2 + self.y**2)
        return norm

    def normalized(self) -> Self:
        """ return a normalized version of the vector
        the initial_point will be the origin (0, 0) and the magnitude will be 1
        :raises ZeroDivisionError if the vector has magnitude zero
        """
        if self.mag == 0:
            raise ZeroDivisionError("Can not normalize a vector of magnitude 0")
        return self / self.mag

    def orthogonal(self, ccw: bool = False) -> Self:
        """ return a normalized vector that points in the (counter)clockwise
        orthogonal direction from this vector
        :argument ccw if True rotate counterclockwise, otherwise clockwise
        :raises ZeroDivisionError if the vector has magnitude zero
        """
        if self.mag == 0:
            raise ZeroDivisionError("Can not normalize a vector of magnitude 0")
        norm = self.normalized()
        if ccw:
            return Vector(norm.y, -norm.x)
        return Vector(-norm.y, norm.x)


ORIGIN = Vector(0, 0)
